@@ -18,7 +18,7 @@
     License along with this library; if not, write to the Free
     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: util.c,v 1.127 2002/02/04 13:46:38 mmclouglin Exp $
+    $Id$
 
 ***************************************************************************/
 #include <assert.h>
@@ -28,13 +28,21 @@
 #include <ctype.h>
 #include <string.h>
 #include <errno.h>
-#ifndef _WIN32
+#ifdef HAVE_UNISTD_H
 #  include <unistd.h>
 #endif
+
 #include "rename.h"
 #include "util.h"
 
-IDL_EXPORT const char *IDL_tree_type_names[] = {
+#ifdef G_OS_WIN32
+#include <fcntl.h>
+#include <direct.h>
+#define popen _popen
+#define pclose _pclose
+#endif
+
+const char *IDL_tree_type_names[] = {
 	"IDLN_NONE",
 	"IDLN_ANY",
 	"IDLN_LIST",
@@ -83,9 +91,9 @@ IDL_EXPORT const char *IDL_tree_type_names[] = {
 	/* IDLN_LAST */
 };
 
-IDL_EXPORT int				__IDL_check_type_casts = FALSE;
+int					__IDL_check_type_casts = FALSE;
 #ifndef HAVE_CPP_PIPE_STDIN
-char *					__IDL_tmp_filename = NULL;
+const char *				__IDL_tmp_filename = NULL;
 #endif
 const char *				__IDL_real_filename = NULL;
 char *					__IDL_cur_filename = NULL;
@@ -218,18 +226,26 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	extern FILE *__IDL_in;
 	FILE *input;
 	char *cmd;
-	size_t cmd_len;
 #ifdef HAVE_CPP_PIPE_STDIN
-	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " - %s%s %s < \"%s\" %s";
-	char *wd = "", *dirend;
+	char *fmt = CPP_PROGRAM " - %s%s %s < \"%s\" %s";
+	char *wd;
 #else
-	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " -I- -I%s %s \"%s\" %s";
-	char *s, *tmpfilename;
+	char *fmt = CPP_PROGRAM " -I%s %s \"%s\" %s";
 	char cwd[2048];
+#ifdef HAVE_SYMLINK
+	char *s, *tmpfilename;
 	gchar *linkto;
+#else
+	const char *tmpfilename;
 #endif
+#endif
+
+#ifndef G_OS_WIN32
 	char *cpperrs = (parse_flags & IDLF_SHOW_CPP_ERRORS) 
 	  ? "" : "2>/dev/null";
+#else
+	char *cpperrs = "";
+#endif
 	GSList *slist;
 	int rv;
 
@@ -240,103 +256,81 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	}
 #endif
 
-	if (!filename ||
-	    !tree ||
-	    (tree == NULL && ns != NULL)) {
+	if (!filename || !tree) {
 		errno = EINVAL;
 		return -1;
 	}
 
-#ifndef NO_ACCESS
+#ifdef HAVE_ACCESS
 	if (access (filename, R_OK))
 		return -1;
 #endif
 
 #ifdef HAVE_CPP_PIPE_STDIN
-	if ((dirend = strrchr (filename, '/'))) {
-		int len = dirend - filename + 1;
-		wd = g_malloc (len);
-		strncpy (wd, filename, len - 1);
-		wd[len - 1] = 0;
-	}
+	wd = g_path_get_dirname (filename);
 
-	cmd_len = (strlen (filename) + (*wd ? 2 : 0) + strlen (wd) +
-		  (cpp_args ? strlen (cpp_args) : 0) +
-		  strlen(cpperrs) +
-		  strlen (fmt) - 8 + 1);
-	cmd = g_malloc (cmd_len);
-	if (!cmd) {
-		errno = ENOMEM;
-		return -1;
-	}
+	cmd = g_strdup_printf (fmt, "-I", wd, cpp_args ? cpp_args : "",
+			       filename, cpperrs);
 
-	g_snprintf (cmd, cmd_len, fmt, *wd ? "-I" : "", wd,
-		    cpp_args ? cpp_args : "", filename, cpperrs);
-
-	if (dirend)
-		g_free (wd);
+	g_free (wd);
 #else
+	if (!getcwd (cwd, sizeof (cwd)))
+		return -1;
+
+#ifdef HAVE_SYMLINK
 	s = tmpnam (NULL);
 	if (s == NULL)
 		return -1;
 
-	if (!getcwd (cwd, sizeof (cwd)))
-		return -1;
-
-	if (*filename == '/') {
+	if (g_path_is_absolute (filename)) {
 		linkto = g_strdup (filename);
 	} else {
-		linkto = g_malloc (strlen (cwd) + strlen (filename) + 2);
-		if (!linkto) {
-			errno = ENOMEM;
-			return -1;
-		}
-		strcpy (linkto, cwd);
-		strcat (linkto, "/");
-		strcat (linkto, filename);
+		linkto = g_strconcat (cwd, "/", filename, NULL);
 	}
 
-	tmpfilename = g_malloc (strlen (s) + 3);
-	if (!tmpfilename) {
-		g_free (linkto);
-		errno = ENOMEM;
-		return -1;
-	}
-	strcpy (tmpfilename, s);
-	strcat (tmpfilename, ".c");
-#ifndef NO_SYMLINK
+	tmpfilename = g_strconcat (s, ".c", NULL);
+
 	if (symlink (linkto, tmpfilename) < 0) {
 		g_free (linkto);
 		g_free (tmpfilename);
 		return -1;
 	}
-#endif
 	g_free (linkto);
-
-	cmd_len = (strlen (tmpfilename) + strlen (cwd) +
-		   (cpp_args ? strlen (cpp_args) : 0) +
-		   strlen(cpperrs) +
-		   strlen (fmt) - 6 + 1);
-	cmd = g_malloc (cmd_len);
-	if (!cmd) {
-		g_free (tmpfilename);
-		errno = ENOMEM;
-		return -1;
-	}
-
-	g_snprintf (cmd, cmd_len, fmt,
-		    cwd, cpp_args ? cpp_args : "", tmpfilename, cpperrs);
+#else
+	tmpfilename = filename;
 #endif
 
-#ifndef NO_POPEN
-	input = popen (cmd, "r");
+	cmd = g_strdup_printf (fmt, cwd, cpp_args ? cpp_args : "",
+			       tmpfilename, cpperrs);
+#endif
+
+       /* Many versions of cpp do evil translating internal
+        * strings, producing bogus output, so clobber LC_ALL */
+	putenv ("LC_ALL=C");
+
+#ifdef HAVE_POPEN
+#if defined (G_OS_WIN32) && !defined (_MSC_VER)
+	if (!(parse_flags & IDLF_SHOW_CPP_ERRORS)) {
+		int save_stderr = dup (2);
+		int null = open ("NUL:", O_WRONLY);
+		dup2 (null, 2);
+		input = popen (cmd, "r");
+		close (2);
+		close (null);
+		dup2 (save_stderr, 2);
+		close (save_stderr);
+	} else
+		input = popen (cmd, "r");
 #else
-	input = fopen (cmd, "r");
+	input = popen (cmd, "r");
+#endif
+#else
+#error Must have popen
 #endif
 	g_free (cmd);
 
 	if (input == NULL || ferror (input)) {
-#ifndef HAVE_CPP_PIPE_STDIN
+#if !defined (HAVE_CPP_PIPE_STDIN) && defined (HAVE_SYMLINK)
 		g_free (tmpfilename);
 #endif
 		return IDL_ERROR;
@@ -364,12 +358,12 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 #ifndef HAVE_CPP_PIPE_STDIN
 	__IDL_tmp_filename = NULL;
 #endif
-#ifndef NO_POPEN
+#ifdef HAVE_POPEN
 	pclose (input);
 #else
 	fclose (input);
 #endif
-#ifndef HAVE_CPP_PIPE_STDIN
+#if !defined (HAVE_CPP_PIPE_STDIN) && defined (HAVE_SYMLINK)
 	unlink (tmpfilename);
 	g_free (tmpfilename);
 #endif
@@ -387,8 +381,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	__IDL_msgcb = NULL;
 
 	if (rv != 0 || !__IDL_is_okay) {
-		if (tree)
-			*tree = NULL;
+		*tree = NULL;
 
 		if (ns)
 			*ns = NULL;
@@ -399,10 +392,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	if (__IDL_flags & IDLF_PREFIX_FILENAME)
 		IDL_ns_prefix (__IDL_root_ns, filename);
 
-	if (tree)
-		*tree = __IDL_root;
-	else
-		IDL_tree_free (__IDL_root);
+	*tree = __IDL_root;
 
 	if (ns)
 		*ns = __IDL_root_ns;
@@ -427,13 +417,11 @@ int IDL_parse_filename_with_input (const char *filename,
 	GSList *slist;
 	int rv;
 
-	if (!filename || !input_cb || !tree ||
-	    (tree == NULL && ns != NULL)) {
+	if (!filename || !input_cb || !tree) {
 		errno = EINVAL;
 		return -1;
 	}
 
-
 	IDL_parse_setup(parse_flags, max_msg_level);
 
 	__IDL_msgcb = msg_cb;
@@ -477,8 +465,7 @@ int IDL_parse_filename_with_input (const char *filename,
 	__IDL_msgcb = NULL;
 
 	if (rv != 0 || !__IDL_is_okay) {
-		if (tree)
-			*tree = NULL;
+		*tree = NULL;
 
 		if (ns)
 			*ns = NULL;
@@ -494,10 +481,7 @@ int IDL_parse_filename_with_input (const char *filename,
 	if (__IDL_flags & IDLF_PREFIX_FILENAME)
 		IDL_ns_prefix (__IDL_root_ns, filename);
 
-	if (tree)
-		*tree = __IDL_root;
-	else
-		IDL_tree_free (__IDL_root);
+	*tree = __IDL_root;
 
 	if (ns)
 		*ns = __IDL_root_ns;
@@ -510,13 +494,13 @@ int IDL_parse_filename_with_input (const char *filename,
 void yyerrorl (const char *s, int ofs)
 {
 	int line = __IDL_cur_line - 1 + ofs;
-	gchar *filename = NULL;
+	gchar *freeme = NULL, *filename = NULL;
 
 	if (__IDL_cur_filename) {
 #ifdef HAVE_CPP_PIPE_STDIN
 		filename = __IDL_cur_filename;
 #else
-		filename = g_basename (__IDL_cur_filename);
+		freeme = filename = g_path_get_basename (__IDL_cur_filename);
 #endif
 	} else
 		line = -1;
@@ -525,8 +509,10 @@ void yyerrorl (const char *s, int ofs)
 	__IDL_is_okay = FALSE;
 
 	/* Errors are counted, even if not printed */
-	if (__IDL_max_msg_level < IDL_ERROR)
+	if (__IDL_max_msg_level < IDL_ERROR) {
+		g_free (freeme);
 		return;
+	}
 
 	if (__IDL_msgcb)
 		(*__IDL_msgcb)(IDL_ERROR, __IDL_nerrors, line, filename, s);
@@ -536,12 +522,13 @@ void yyerrorl (const char *s, int ofs)
 		else
 			fprintf (stderr, "Error: %s\n", s);
 	}
+	g_free (freeme);
 }
 
 void yywarningl (int level, const char *s, int ofs)
 {
 	int line = __IDL_cur_line - 1 + ofs;
-	gchar *filename = NULL;
+	gchar *freeme = NULL, *filename = NULL;
 
 	/* Unprinted warnings are not counted */
 	if (__IDL_max_msg_level < level)
@@ -551,7 +538,7 @@ void yywarningl (int level, const char *s, int ofs)
 #ifdef HAVE_CPP_PIPE_STDIN
 		filename = __IDL_cur_filename;
 #else
-		filename = g_basename (__IDL_cur_filename);
+		freeme = filename = g_path_get_basename (__IDL_cur_filename);
 #endif
 	} else
 		line = -1;
@@ -566,6 +553,7 @@ void yywarningl (int level, const char *s, int ofs)
 		else
 			fprintf (stderr, "Warning: %s\n", s);
 	}
+	g_free (freeme);
 }
 
 void yyerror (const char *s)
@@ -1887,10 +1875,18 @@ static void IDL_tree_free_real (IDL_tree p)
 		g_free (IDL_STRING (p).value);
 		break;
 
+	case IDLN_WIDE_STRING:
+		g_free (IDL_WIDE_STRING (p).value);
+		break;
+
 	case IDLN_CHAR:
 		g_free (IDL_CHAR (p).value);
 		break;
 
+	case IDLN_WIDE_CHAR:
+		g_free (IDL_WIDE_CHAR (p).value);
+		break;
+
 	case IDLN_IDENT:
 		g_free (IDL_IDENT (p).str);
 		g_free (IDL_IDENT_REPO_ID (p));
@@ -1955,9 +1951,12 @@ void IDL_tree_free (IDL_tree p)
 	case IDLN_TYPE_TYPECODE:
 	case IDLN_FIXED:
 	case IDLN_STRING:
+	case IDLN_WIDE_STRING:
 	case IDLN_CHAR:
+	case IDLN_WIDE_CHAR:
 	case IDLN_IDENT:
 	case IDLN_CODEFRAG:
+	case IDLN_SRCFILE:
 		__IDL_tree_free (p);
 		break;
 
@@ -2605,7 +2604,7 @@ static gboolean IDL_output_delim_pre (IDL_tree_func_data *tfd, IDL_output_delim_
 {
 	if (IDL_output_delim_match (tfd, delim)) {
 		if (delim->hit)
-			dataf (delim->data, delim->delim);
+			dataf (delim->data, "%s", delim->delim);
 		else
 			delim->hit = TRUE;
 		return delim->pre_func
@@ -2767,9 +2766,11 @@ static gboolean IDL_emit_IDL_ident_real (IDL_tree_func_data *tfd, IDL_output_dat
 	assert (IDL_NODE_TYPE (tfd->tree) == IDLN_IDENT);
 
 	if (!up_real || data->flags & IDLF_OUTPUT_NO_QUALIFY_IDENTS) {
+		/* TODO: If the IDENT is also a keyword, escape it by
+		   prepending an underscore. */
 		dataf (data, "%s", IDL_IDENT (tfd->tree).str);
 	} else {
-		if ( up_path == 0 ) {
+		if (up_path == NULL) {
 		    levels = 0;
 		} else {
 		    /* Determine minimal required levels of scoping */
@@ -3359,7 +3360,7 @@ struct IDL_emit_node {
 	IDL_tree_func post;
 };
 
-static const struct IDL_emit_node * const IDL_get_IDL_emission_table (void)
+static struct IDL_emit_node * IDL_get_IDL_emission_table (void)
 {
 	static gboolean initialized = FALSE;
 	static struct IDL_emit_node table[IDLN_LAST];
@@ -3527,7 +3528,7 @@ static gboolean
 contains_node_walker (IDL_tree_func_data *tfd, gpointer data) {
 	ContainsNodeWalkerInfo	*info = data;
 	/* skip root! */
-	if ( tfd->up!=0 && tfd->tree == info->searchNode ) {
+	if (tfd->up != NULL && tfd->tree == info->searchNode) {
 		info->found = TRUE;
 		return FALSE;
 	}
@@ -3662,7 +3663,7 @@ IDL_tree_is_recursive (IDL_tree tree, gpointer dummy)
 	info.ident_list = NULL;
 	info.recursive  = FALSE;
 
-	IDL_tree_walk2 (tree, 0, IDL_WalkF_TypespecOnly, 
+	IDL_tree_walk2 (tree, NULL, IDL_WalkF_TypespecOnly, 
 			IDL_tree_is_recursive_walker_pre, 
 			IDL_tree_is_recursive_walker_post, 
 			&info);
